class IncompatibleAttribute(Exception):
    pass


class IncompatibleDataException(Exception):
    pass


class UndefinedROI(Exception):
    pass


class InvalidSubscriber(Exception):
    pass


class InvalidMessage(Exception):
    pass
